/*
 * Apple built-in iSight firmware loader
 *
 * Copyright © 2006 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright © 2007 Étienne Bersac <bersace@gmail.com>
 *
 *
 * Special thanks to Johannes Berg <johannes@sipsolutions.net> for helping
 * to find out how to load the firmware; see his website on
 * http://johannes.sipsolutions.net/MacBook/iSight for details.
 * Basic structure of firmware data: <len:2><off:2><data:len>, where
 * the $data of size $len is to be put at position $off in the device.
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>


#define	ift_message(args...)	syslog(LOG_INFO, args)
#define	ift_warning(args...)	syslog(LOG_WARNING, args)
#define	ift_error(args...)	({syslog(LOG_ERR, args); exit(-1);})


#include <usb.h>
#define TIMEOUT 300

/* OPTIONS */
static gchar		*firmware	= "isight.fw";
static const gchar	*USB_BUS	= NULL;
static const gchar	*USB_DEV	= NULL;

static GOptionEntry entries[] = {
	{ "firmware", 'f',
	  0,
	  G_OPTION_ARG_FILENAME, &firmware,
	  N_("Path to the firmware."),
	  NULL
	},
	{ "usb-bus", 'b',
	  0,
	  G_OPTION_ARG_STRING, &USB_BUS,
	  N_("USB bus id"),
	  NULL
	},
	{ "usb-dev", 'd',
	  0,
	  G_OPTION_ARG_STRING, &USB_DEV,
	  N_("USB device id"),
	  NULL
	},
	{ NULL }
};

static int
load_firmware(struct usb_device *dev, char *firmware)
{
	int fd, len, req, llen, res, ret = -1;
	unsigned char data[4], rdata[1024], *ptr;
	usb_dev_handle *h;

	/* load firmware */
	if (!(h = usb_open(dev))) {
		ift_warning(_("Failed to open device"));
		return -1;
	}

	if ((fd = g_open(firmware, O_RDONLY)) == -1) {
		ift_warning(_("Failed to open firmware"));
		return -1;
	}

	if ((res = usb_control_msg(h, 0x40, 0xA0, 0xe600, 0,
				   "\1", 1, TIMEOUT)) != 1) {
		ift_warning(_("Failed to init firmware loading"));
		close(fd);
		return -1;
	}
	while (1) {
		/* read length and offset */
		if ((len = read(fd, data, 4)) != 4) {
			if (len == 0) {
				break;
				goto end;
			} else {
				ift_warning(_("Reading firmware header chunk "
					      "failed"));
				goto end;
			}
		}

		len =(data[0] << 8) | data[1];
		req = (data[2] << 8) | data[3];

		/* read data */
		if (len < 0 || len >= 1024) {
			ift_warning(_("Invalid firmware data_length %d, "
				      "load aborted\n"),
				    len);
			goto end;
		} else if (read(fd, rdata, len) != len) {
			ift_warning(_("Failed to read firmware data"));
			goto end;
		}

		/* upload to usb bus */
		for (ptr = rdata; len > 0; req += 50, ptr += 50) {
			llen = len > 50 ? 50 : len;
			len -= llen;

			if ((res = usb_control_msg(h, 0x40, 0xA0, req, 0,
						   (char *) ptr, llen,
						   TIMEOUT)) != llen) {
				ift_warning(_("Firmware load req=0x%x failed: %s"),
					    req, strerror(errno));
				goto end;
			}
		}
	}

	ret = 0;

 end:
	if ((res = usb_control_msg(h, 0x40, 0xA0, 0xe600, 0,
				   "\0", 1, TIMEOUT)) != 1) {
		ift_warning(_("Failed to close firmware loading"));
		ret = -1;
	}

	close(fd);
	usb_close(h);

	return ret;
}

static struct usb_device*
find_usb_device(const gchar *bus_id, const gchar *dev_id)
{
	struct usb_bus *bi, *bus = NULL;
	struct usb_device *di, *dev = NULL;

	for (bi = usb_busses; bi != NULL; bi = bi->next) {
		if (g_str_equal(bi->dirname, bus_id)) {
			bus = bi;
			break;
		}
	}

	if (!bus) {
		ift_warning(_("USB Bus %s not found"), bus_id);
		return NULL;
	}

	for (di = bus->devices; di != NULL; di = di->next) {
		if (g_str_equal(di->filename, dev_id)) {
			dev = di;
			break;
		}
	}

	if (!dev)
		ift_warning(_("USB device %s not found in bus %s"),
			    dev_id, bus_id);

	return dev;
}

int
main(int argc, char *argv[])
{
	GOptionContext *context;
	GError *error = NULL;
	struct usb_device *dev;

	openlog("ift-load", LOG_PERROR, LOG_USER);

#if ENABLE_NLS
	/* ??????? */
	setlocale(LC_ALL, "");
	bindtextdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
	textdomain(GETTEXT_PACKAGE);
#endif

	gchar *appname = _("built-in iSight firmware loader");

	g_set_application_name(appname);
	g_set_prgname("ift-load");

	/* translators: this is the first --help output line */
	context = g_option_context_new(g_strdup_printf(" - %s",
						       appname));

	/* tip at the end of --help output */
	gchar *desc = _("Use ift-extract to extract firmware.");
	g_option_context_set_description(context, desc);
	/* translators: this is shown at the end of --help output */
	g_option_context_add_main_entries(context, entries, GETTEXT_PACKAGE);
	if (!g_option_context_parse(context, &argc, &argv, &error)) {
		fprintf(stderr, _("Error: %s\n"),
			error->message); /* comment traduire ? */
		fprintf(stderr, g_option_context_get_help(context,
							  FALSE, NULL));
	}

	if (g_access(firmware, R_OK)) {
		ift_error(_("Unable to read firmware %s."), firmware);
	}

	/* DEVICE IDS */
	/* this is ubuntu/debian specific */
	USB_BUS = USB_BUS ? USB_BUS : g_getenv("USB_BUS");
	USB_DEV = USB_DEV ? USB_DEV : g_getenv("USB_DEV");

	/* this should work everywhere. */
	if (!USB_BUS || !USB_DEV) {
		const gchar *device = g_getenv("DEVICE");
		if (device) {
			gchar **parts = g_strsplit(device, "/", 6);
			USB_BUS = g_strdup(parts[4]);
			USB_DEV = g_strdup(parts[5]);
			g_strfreev(parts);
		}
	}

	if (!USB_BUS)
		ift_error(_("No USB bus id specified"));

	if (!USB_DEV)
		ift_error(_("No USB device id specified"));

	/* init usb */
	usb_init();
	if (usb_find_busses() == 0) {
		ift_error(_("No USB busses found"));
	} else if (usb_find_devices() == 0) {
		ift_error(_("No USB devices found"));
	}


	dev = find_usb_device(USB_BUS, USB_DEV);
	if (!dev)
		ift_error(_("Device %s:%s not found"), USB_BUS, USB_DEV);

	if (load_firmware(dev, firmware) == -1)
		ift_error(_("Failed to upload firmware to %s:%s (%x:%x)"),
			  USB_BUS, USB_DEV,
			  dev->descriptor.idVendor,
			  dev->descriptor.idProduct);
	else
		ift_message(_("Firmware loaded succesfully to %s:%s"),
			    USB_BUS, USB_DEV);

	closelog();

	return 0;
}

